@@ -1,12 +1,12 @@
 #
-# Copyright (C) 2015 J. Maslak
+# Copyright (C) 2015-2017 Joelle Maslak
 # All Rights Reserved - See License
 #
 
+package Parallel::WorkUnit;
+$Parallel::WorkUnit::VERSION = '1.012';
 use v5.8;
 
-package Parallel::WorkUnit;
-$Parallel::WorkUnit::VERSION = '1.010';
 # ABSTRACT: Provide easy-to-use forking with ability to pass back data
 
 use strict;
@@ -14,7 +14,10 @@ use warnings;
 use autodie;
 
 use Try::Tiny;
-my $do_thread = eval 'use threads qw//; 1' if $^O eq 'MSWin32';
+my $do_thread;
+## no critic (BuiltinFunctions::ProhibitStringyEval)
+$do_thread = eval 'use threads qw//; 1' if $^O eq 'MSWin32';
+## critic
 if ($do_thread) { eval 'use Thread::Queue;'; }
 
 use Carp;
@@ -22,12 +25,24 @@ use Carp;
 use IO::Pipe;
 use IO::Select;
 use Moose;
+use Moose::Util::TypeConstraints;
 use POSIX ':sys_wait_h';
 use Storable;
 
 use namespace::autoclean;
 
 
+subtype 'Parallel::WorkUnit::PositiveInt', as 'Int',
+  where { $_ > 0 },
+  message { "The number you provided, $_, was not a positive number" };
+
+
+has 'max_children' => (
+    is      => 'rw',
+    isa     => 'Maybe[Parallel::WorkUnit::PositiveInt]',
+    default => 5,
+    trigger => sub { $_[0]->_start_queued_children() },
+);
 
 has '_subprocs' => (
     is       => 'rw',
@@ -49,6 +64,15 @@ has '_count' => (
     default  => 1
 );
 
+# Children queued
+has '_queued_children' => (
+    is       => 'rw',
+    isa      => 'ArrayRef[ArrayRef[Coderef]]',
+    init_arg => undef,
+    default  => sub { [] },
+);
+
+
 sub BUILD {
     my $self = shift;
 
@@ -57,6 +81,8 @@ sub BUILD {
     if ($do_thread) {
         $self->_queue( Thread::Queue->new() );
     }
+
+    return;
 }
 
 
@@ -66,12 +92,12 @@ sub async {
 
     my $pipe = IO::Pipe->new();
 
-    my ($pid, $thr);
+    my ( $pid, $thr );
     if ($do_thread) {
         $pid = $self->_count();
-        $self->_count($pid + 1);
+        $self->_count( $pid + 1 );
 
-        $thr = threads->create( sub { $self->_child($sub, $pipe, $pid); } );
+        $thr = threads->create( sub { $self->_child( $sub, $pipe, $pid ); } );
         if ( !defined($thr) ) { die "thread creation failed: $!"; }
     } else {
         $pid = fork();
@@ -91,13 +117,13 @@ sub async {
         return $pid;
 
     } else {
-        $self->_child($sub, $pipe, undef);
+        return $self->_child( $sub, $pipe, undef );
     }
 }
 
 sub _child {
-    if (scalar(@_) != 4) { confess 'invalid call'; }
-    my ($self, $sub, $pipe, $pid) = @_;
+    if ( scalar(@_) != 4 ) { confess 'invalid call'; }
+    my ( $self, $sub, $pipe, $pid ) = @_;
 
     # We are in the child process
     $pipe->writer();
@@ -106,7 +132,8 @@ sub _child {
     try {
         my $result = $sub->();
         $self->_send_result( $pipe, $result, $pid );
-    } catch {
+    }
+    catch {
         $self->_send_error( $pipe, $_, $pid );
     };
 
@@ -131,9 +158,9 @@ sub waitall {
 sub waitone {
     if ( $#_ != 0 ) { confess 'invalid call'; }
     my ($self) = @_;
-    
+
     my $sp = $self->_subprocs();
-    if ( !keys(%$sp) ) { return undef; }
+    if ( !keys(%$sp) ) { return; }
 
     if ($do_thread) {
         # On Windows
@@ -144,6 +171,9 @@ sub waitone {
         $self->_read_result($child);
         $thr->join();
 
+        # Start queued children, if needed
+        $self->_start_queued_children();
+
         return 1;
     } else {
         # On everything but Windows
@@ -155,7 +185,7 @@ sub waitone {
 
         foreach my $fh (@ready) {
             foreach my $child ( keys(%$sp) ) {
-                if ( defined($fh->fileno())) {
+                if ( defined( $fh->fileno() ) ) {
                     if ( $fh->fileno() == $sp->{$child}{fh}->fileno() ) {
                         my $thr = $self->_subprocs()->{$child}{thread};
                         $self->_read_result($child);
@@ -163,10 +193,13 @@ sub waitone {
                         if ($do_thread) {
                             $thr->join();
                         } else {
-                            waitpid($child, 0);
+                            waitpid( $child, 0 );
                         }
 
-                        return 1;  # We don't want to read more than one!
+                        # Start queued children, if needed
+                        $self->_start_queued_children();
+
+                        return 1;    # We don't want to read more than one!
                     }
                 }
             }
@@ -174,10 +207,11 @@ sub waitone {
     }
 
     # We should never get here
-    return undef;
+    return;
 }
 
 
+## no critic ('Subroutines::ProhibitBuiltinHomonyms')
 sub wait {
     if ( $#_ != 1 ) { confess 'invalid call'; }
     my ( $self, $pid ) = @_;
@@ -189,39 +223,49 @@ sub wait {
         return;
     }
 
-    my $thr = $self->_subprocs()->{$pid}{thread};
+    my $thr    = $self->_subprocs()->{$pid}{thread};
     my $result = $self->_read_result($pid);
 
     if ($do_thread) {
         $thr->join();
     } else {
-        waitpid($pid, 0);
+        waitpid( $pid, 0 );
     }
 
     return $result;
 }
+## use critic
 
 
 sub count {
     if ( $#_ != 0 ) { confess 'invalid call'; }
     my ($self) = @_;
-    
+
     my $sp = $self->_subprocs();
-    return scalar(keys %$sp);
+    return scalar( keys %$sp );
+}
+
+
+sub queue {
+    if ( $#_ != 2 ) { confess 'invalid call'; }
+    my ( $self, $sub, $callback ) = @_;
+
+    push @{ $self->_queued_children }, [ $sub, $callback ];
+    return $self->_start_queued_children();
 }
 
 sub _send_result {
     if ( $#_ != 3 ) { confess 'invalid call'; }
     my ( $self, $fh, $msg, $pid ) = @_;
 
-    $self->_send( $fh, 'RESULT', $msg, $pid );
+    return $self->_send( $fh, 'RESULT', $msg, $pid );
 }
 
 sub _send_error {
     if ( $#_ != 3 ) { confess 'invalid call'; }
     my ( $self, $fh, $err, $pid ) = @_;
 
-    $self->_send( $fh, 'ERROR', $err, $pid );
+    return $self->_send( $fh, 'ERROR', $err, $pid );
 }
 
 sub _send {
@@ -230,7 +274,7 @@ sub _send {
 
     my $msg = Storable::freeze( \$data );
 
-    if (!defined($msg)) {
+    if ( !defined($msg) ) {
         die 'freeze() returned undef for child return value';
     }
 
@@ -241,14 +285,15 @@ sub _send {
     $fh->write($type);
     $fh->write("\n");
 
-    $fh->write(length($msg));
+    $fh->write( length($msg) );
     $fh->write("\n");
 
-    binmode($fh, ':raw');
+    binmode( $fh, ':raw' );
 
     $fh->write($msg);
 
     $fh->close();
+    return;
 }
 
 sub _read_result {
@@ -269,7 +314,6 @@ sub _read_result {
 
     my $result = '';
 
-    my $part;
     my $ret = 1;
     while ( defined($ret) && ( length($result) < $size ) ) {
         my $s = $size - length($result);
@@ -282,7 +326,7 @@ sub _read_result {
     my $data = ${ Storable::thaw($result) };
 
     my $caller = $self->_subprocs()->{$child}{caller};
-    my $thr = $self->_subprocs()->{$child}{thread};
+    my $thr    = $self->_subprocs()->{$child}{thread};
     delete $self->_subprocs()->{$child};
     $fh->close();
 
@@ -290,9 +334,38 @@ sub _read_result {
         $cinfo->{callback}->($data);
     } else {
         if ($do_thread) { $thr->join(); }
-        die("Child (created at " . $caller->[1] . " line " . $caller->[2] .
-            ") died with error: $data");
+        die(    "Child (created at "
+              . $caller->[1]
+              . " line "
+              . $caller->[2]
+              . ") died with error: $data" );
     }
+
+    return;
+}
+
+# Start queued children, if possible.
+# Returns 1 if children were started, undef otherwise
+sub _start_queued_children() {
+    if ( $#_ != 0 ) { confess 'invalid call' }
+    my ($self) = @_;
+
+    if ( !( @{ $self->_queued_children } ) ) { return; }
+
+    # Can we start a queued process?
+    while ( scalar @{ $self->_queued_children } ) {
+        if ( ( !defined( $self->max_children ) ) || ( $self->count < $self->max_children ) ) {
+            # Start queued child
+            my $ele = shift @{ $self->_queued_children };
+            $self->async( $ele->[0], $ele->[1] );
+        } else {
+            # Can't unqueue
+            return;
+        }
+    }
+
+    # We started at least one process
+    return 1;
 }
 
 __PACKAGE__->meta->make_immutable;
@@ -311,7 +384,7 @@ Parallel::WorkUnit - Provide easy-to-use forking with ability to pass back data
 
 =head1 VERSION
 
-version 1.010
+version 1.012
 
 =head1 SYNOPSIS
 
@@ -320,6 +393,10 @@ version 1.010
 
   $wu->waitall();
 
+  $wu->max_children(5);
+  $wu->queue( sub { ... }, \&callback );
+  $wu->waitall();
+
 =head1 DESCRIPTION
 
 This is a very simple forking implementation of parallelism, with the
@@ -332,11 +409,31 @@ of data back to the parent process.
 There are many other Parallel::* applications in CPAN - it would be worth
 any developer's time to look through those and choose the best one.
 
+=head1 ATTRIBUTES
+
+=head2 max_children
+
+  $wu->max_children(5);
+  $wu->max_children(undef);
+
+  say "Max number of children: " . $wu->max_children();
+
+If set to a value other than undef, limits the number of outstanding
+queue children (created by the C<queue()> method) that can be executing
+at any given time.  This defaults to 5.
+
+Calling without any parameters will return the number of children.
+
 =head1 METHODS
 
 =head2 new
 
-Create a new workunit class.
+Create a new workunit class.  Optionally, takes a list that corresponds
+to a hashref, in the form of key and value.  This accepts the key
+C<max_children>, which, if present (and not undef) will limit the
+number of spawned subprocesses that can be active when using the
+C<queue()> method.  Defaults to 5.  See the C<max_children> method
+for additional information.
 
 =head2 async( sub { ... }, \&callback )
 
@@ -361,6 +458,11 @@ die (inside the C<waitall()> method).
 The PID of the child is returned to the parent process when
 this method is executed.
 
+The C<max_children> attribute is not examined in this method - you
+can spawn a new child regardless of the number of children already
+spawned. However, you children started with this method still count
+against the limit used by C<queue()>.
+
 Note: on Windows with threaded Perl, threads instead of forks are used.
 See C<thread> for the caveats that apply.  The PID returned is instead
 a meaningless (outside of this module) counter, not associated with any
@@ -402,6 +504,19 @@ This method returns the number of currently outstanding
 threads (in either a running state or a waiting to send their
 output).
 
+=head2 queue( sub { ... }, \&callback )
+
+Spawns work on a new forked process, doing so immediately if less
+than C<max_children> are running.  If there are already
+C<max_children> are running, this will run the process once a slot
+becomes available.
+
+This method should be treated as nearly identical to C<async()>,
+with the only difference being the above behavior (limiting to
+C<max_children>) and not returning a PID.  Instead, a value of 1
+is returned if the process is immediately started, C<undef>
+otherwise.
+
 =head1 BUGS
 
 Windows doesn't do C<fork()>, but emulates it with threads.  As a result,
@@ -414,11 +529,11 @@ threaded Perl.
 
 =head1 AUTHOR
 
-J. Maslak <jmaslak@antelope.net>
+Joelle Maslak <jmaslak@antelope.net>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2016 by J. Maslak.
+This software is copyright (c) 2017 by Joelle Maslak.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
